/*
 * ie.c  -- IE depend
 *
 * $Id: ie.c,v 1.8 2006/02/03 11:11:49 hos Exp $
 *
 */


#include "main.h"
#include <oleacc.h>


static HRESULT ie_get_doc_body(wnd_ctx_t *ctx, IDispatch **body)
{
    BSTR compat_mode = NULL;
    HRESULT hres;

    hres = get_property_str(ctx->doc, L"compatMode", &compat_mode);
    if(SUCCEEDED(hres) &&
       compat_mode != NULL && wcscmp(compat_mode, L"CSS1Compat") == 0) {
        hres = get_property_dp(ctx->doc, L"documentElement", body);
    } else {
        hres = get_property_dp(ctx->doc, L"body", body);
    }
    SysFreeString(compat_mode);

    return hres;
}


static HRESULT ie_selected_range(wnd_ctx_t *ctx, IDispatch **sel_range)
{
    IDispatch *range;
    BSTR str = NULL;
    HRESULT hres;
    int use_sel;
    LPOLESTR doc_activeelem_tag[] = {L"activeElement", L"tagName", NULL};
    LPOLESTR doc_selection_range[] = {L"selection", L"createRange", NULL};

    hres = get_property_strv(ctx->doc, doc_activeelem_tag, &str);
    use_sel = (FAILED(hres) || wcscmp(str, L"INPUT") != 0);
    SysFreeString(str);
    str = NULL;

    if(use_sel) {
        hres = get_property_dpv(ctx->doc, doc_selection_range, &range);
        if(FAILED(hres)) {
            return hres;
        }

        hres = get_property_str(range, L"text", &str);
    } else {
        hres = get_property_dpv(ctx->doc, ctx->doc_all_range, &range);
        if(FAILED(hres)) {
            return hres;
        }
    }

    if((! use_sel) || FAILED(hres) || str == NULL || str[0] == 0) {
        long t = 0, l = 0, w = 0, h = 0, x, y;
        VARIANTARG args[2];
        IDispatch *body;

        hres = ie_get_doc_body(ctx, &body);
        if(SUCCEEDED(hres)) {
            get_property_long(body, L"clientTop", &t);
            get_property_long(body, L"clientLeft", &l);
            get_property_long(body, L"clientWidth", &w);
            get_property_long(body, L"clientHeight", &h);
            IDispatch_Release(body);

            if(ctx->dir > 0) {
                x = l + 5;
                y = t + 5;
            } else {
                x = l + w - 5;
                y = t + h - 5;
            }

            args[0].vt = VT_I4;
            args[0].lVal = y;
            args[1].vt = VT_I4;
            args[1].lVal = x;

            call_method_s(range, L"moveToPoint", args, 2, NULL);
        }
    }
    SysFreeString(str);

    *sel_range = range;

    return S_OK;
}

static HRESULT ie_range_find(IDispatch *range, wnd_ctx_t *ctx, BSTR text,
                             VARIANT_BOOL *res)
{
    VARIANTARG args[2];
    VARIANT var;
    HRESULT hres;

    args[0].vt = VT_I4;
    args[0].lVal = ctx->dir;
    args[1].vt = VT_BSTR;
    args[1].bstrVal = text;

    hres = call_method_s(range, L"findText", args, 2, &var);
    if(FAILED(hres)) {
        return hres;
    }
    if(var.vt != VT_BOOL) {
        return E_FAIL;
    }

    *res = var.boolVal;

    return S_OK;
}

static HRESULT ie_range_move_point(IDispatch *search_range,
                                   IDispatch *sel_range,
                                   int dir, int offset)
{
    VARIANTARG args[2];
    BSTR str;
    HRESULT hres;

    args[0].vt = VT_DISPATCH;
    args[0].pdispVal = sel_range;
    str = SysAllocString(dir > 0 ? L"StartToStart" : L"EndToEnd");
    args[1].vt = VT_BSTR;
    args[1].bstrVal = str;
    hres = call_method_s(search_range, L"setEndPoint", args, 2, NULL);
    SysFreeString(str);
    if(FAILED(hres)) {
        return hres;
    }

    args[0].vt = VT_I4;
    args[0].lVal = offset;
    str = SysAllocString(L"character");
    args[1].vt = VT_BSTR;
    args[1].bstrVal = str;
    hres = call_method_s(search_range, (dir > 0 ? L"moveStart" : L"moveEnd"),
                         args, 2, NULL);
    SysFreeString(str);
    if(FAILED(hres)) {
        return hres;
    }

    return S_OK;
}

static HRESULT ie_range_select(wnd_ctx_t *ctx, IDispatch *range)
{
    long offset_top, bounding_height, scroll_top, client_height;
    IDispatch *body;
    HRESULT hres;

    hres = get_property_long(range, L"offsetTop", &offset_top);
    if(FAILED(hres)) {
        return hres;
    }

    hres = get_property_long(range, L"boundingHeight", &bounding_height);
    if(FAILED(hres)) {
        return hres;
    }

    hres = ie_get_doc_body(ctx, &body);
    if(FAILED(hres)) {
        return hres;
    }

    hres = get_property_long(body, L"scrollTop", &scroll_top);
    if(FAILED(hres)) {
        IDispatch_Release(body);
        return hres;
    }

    hres = get_property_long(body, L"clientHeight", &client_height);
    if(FAILED(hres)) {
        IDispatch_Release(body);
        return hres;
    }

    if(offset_top < 0 || offset_top + bounding_height > client_height) {
        scroll_top += offset_top - client_height / 2;

        hres = put_property_long(body, L"scrollTop", scroll_top);
        IDispatch_Release(body);
        if(FAILED(hres)) {
            return hres;
        }
    }

    hres = call_method_s(range, L"select", NULL, 0, NULL);
    if(FAILED(hres)) {
        return hres;
    }

    return S_OK;
}

static HRESULT ie_range_action(wnd_ctx_t *ctx, IDispatch *range)
{
    LPOLESTR parent_click[] = {L"parentElement", L"click", NULL};
    HRESULT hres;

    hres = get_property_v(range, parent_click, NULL);
    if(FAILED(hres)) {
        return hres;
    }

    return S_OK;
}


static BOOL CALLBACK enum_html_doc_wnd(HWND hwnd, LPARAM param)
{
    TCHAR buf[256];

    if(GetClassName(hwnd, buf, 256) == 0) {
        return TRUE;
    }

    if(_tcscmp(buf, _T("Internet Explorer_Server")) != 0) {
        return TRUE;
    }

    *((HWND *)param) = hwnd;
    return FALSE;
}

static HRESULT get_ie_active_doc(IDispatch *doc, IDispatch **active_doc)
{
    IDispatch *inner_doc;
    HRESULT hres;

    inner_doc = doc;
    IDispatch_AddRef(inner_doc);

    while(1) {
        IDispatch *elem;
        BSTR str;

        hres = get_property_dp(inner_doc, L"activeElement", &elem);
        if(FAILED(hres)) {
            IDispatch_Release(inner_doc);
            return hres;
        }

        hres = get_property_str(elem, L"tagName", &str);
        if(FAILED(hres)) {
            IDispatch_Release(inner_doc);
            IDispatch_Release(elem);
            return hres;
        }

        if(str == NULL || wcsstr(str, L"FRAME") == NULL) {
            SysFreeString(str);
            IDispatch_Release(elem);
            break;
        }

        SysFreeString(str);
        IDispatch_Release(inner_doc);

        {
            LPOLESTR props[] = {L"contentWindow", L"document", NULL};
            hres = get_property_dpv(elem, props, &inner_doc);
        }
        IDispatch_Release(elem);
        if(FAILED(hres)) {
            return hres;
        }
    }

    *active_doc = inner_doc;

    return S_OK;
}

static HRESULT get_ie_doc(HWND hwnd, IDispatch **doc)
{
    IDispatch *top_doc;
    HWND fg_hwnd, doc_hwnd;
    UINT getobj_msg;
    LRESULT lret, res;
    HRESULT hres;

    fg_hwnd = hwnd;

    doc_hwnd = NULL;
    EnumChildWindows(fg_hwnd, enum_html_doc_wnd, (LPARAM)&doc_hwnd);
    if(doc_hwnd == NULL) {
        return E_FAIL;
    }

    getobj_msg = RegisterWindowMessage(_T("WM_HTML_GETOBJECT"));
    if(getobj_msg == 0) {
        return E_FAIL;
    }

    lret = SendMessageTimeout(doc_hwnd, getobj_msg, 0, 0,
                              SMTO_ABORTIFHUNG, 1000,
                              (PDWORD_PTR)&res);
    if(lret == 0) {
        return E_FAIL;
    }

    hres = ObjectFromLresult(res, &IID_IDispatch, 0,
                             (void **)(void *)&top_doc);
    if(FAILED(hres)) {
        return hres;
    }

    hres = get_ie_active_doc(top_doc, doc);
    if(FAILED(hres)) {
        return hres;
    }

    return S_OK;
}


static LPOLESTR range_text[] = {L"text", NULL};
static LPOLESTR range_dup[] = {L"duplicate", NULL};
static LPOLESTR doc_all_range[] = {L"body", L"createTextRange", NULL};

static HRESULT ie_init_context(wnd_ctx_t *ctx)
{
    ctx->range_text = range_text;
    ctx->range_dup = range_dup;
    ctx->doc_all_range = doc_all_range;

    ctx->selected_range = ie_selected_range;
    ctx->range_find_proc = ie_range_find;
    ctx->range_move_point = ie_range_move_point;
    ctx->range_select = ie_range_select;
    ctx->range_action = ie_range_action;

    ctx->get_doc = get_ie_doc;

    return S_OK;
}


DECLARE_FUNCTIONS(ie)
